/**
 * Copyright (c) 2011-2012 EBM WebSourcing, 2012-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.junit.impl;

import static org.ow2.petals.commons.log.PetalsExecutionContext.FLOW_INSTANCE_ID_PROPERTY_NAME;
import static org.ow2.petals.commons.log.PetalsExecutionContext.FLOW_PREVIOUS_STEP_ID_PROPERTY_NAME;
import static org.ow2.petals.commons.log.PetalsExecutionContext.FLOW_STEP_ID_PROPERTY_NAME;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.Collection;
import java.util.Properties;
import java.util.concurrent.Semaphore;
import java.util.jar.JarFile;

import javax.jbi.JBIException;
import javax.jbi.component.InstallationContext;
import javax.jbi.management.DeploymentException;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.apache.commons.io.FileUtils;
import org.junit.Assert;
import org.ow2.easywsdl.extensions.wsdl4complexwsdl.api.Description;
import org.ow2.petals.commons.log.FlowAttributes;
import org.ow2.petals.commons.log.FlowAttributesExchangeHelper;
import org.ow2.petals.commons.log.Level;
import org.ow2.petals.commons.log.PetalsExecutionContext;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.JBIBootstrap;
import org.ow2.petals.component.framework.api.Message.MEPConstants;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.CDKJBIDescriptorBuilder;
import org.ow2.petals.component.framework.jbidescriptor.generated.Jbi;
import org.ow2.petals.component.framework.junit.Message;
import org.ow2.petals.component.framework.junit.RequestMessage;
import org.ow2.petals.component.framework.junit.ResponseMessage;
import org.ow2.petals.component.framework.junit.impl.ServiceConfiguration.ServiceType;
import org.ow2.petals.component.framework.junit.impl.message.AbstractWrappedToComponentMessage;
import org.ow2.petals.component.framework.junit.impl.message.WrappedRequestFromConsumerMessage;
import org.ow2.petals.component.framework.junit.impl.message.WrappedResponseFromProviderMessage;
import org.ow2.petals.component.framework.junit.impl.mock.MockComponentContext;
import org.ow2.petals.component.framework.junit.impl.mock.MockInstallationContext;
import org.ow2.petals.component.framework.logger.AbstractFlowLogData;
import org.ow2.petals.component.framework.logger.ProvideFlowStepBeginLogData;
import org.ow2.petals.component.framework.logger.ProvideFlowStepEndLogData;
import org.ow2.petals.component.framework.logger.ProvideFlowStepFailureLogData;
import org.ow2.petals.component.framework.message.ExchangeImpl;
import org.ow2.petals.component.framework.process.JBIProcessorManager;
import org.ow2.petals.component.framework.process.JBIProcessorManagerObserver;
import org.ow2.petals.component.framework.su.AbstractServiceUnitManager;
import org.ow2.petals.component.framework.su.ServiceUnitDataHandler;
import org.ow2.petals.jbi.descriptor.JBIDescriptorException;
import org.ow2.petals.jbi.messaging.exchange.PetalsMessageExchange;

import com.ebmwebsourcing.easycommons.io.IOHelper;
import com.ebmwebsourcing.easycommons.io.ZipHelper;
import com.ebmwebsourcing.easycommons.lang.StringHelper;
import com.ebmwebsourcing.easycommons.lang.UncheckedException;
import com.ebmwebsourcing.easycommons.thread.ExecutionContext;

/**
 * @author Marc Jambert - EBM WebSourcing
 */
public class ComponentInstanceImpl {

    private enum ComponentState {
        NOT_INSTALLED, STOPPED, STARTED;
    }

    private static final String JBI_DIR_NAME = "jbi";

    private final File componentRepositoryRoot;

    private final ComponentConfiguration componentConfiguration;

    private final Jbi jbi;

    private final AbstractComponent component;

    private MockComponentContext mockComponentContext;

    private ComponentState componentState = ComponentState.NOT_INSTALLED;

    public ComponentInstanceImpl(final ComponentConfiguration componentConfiguration, final File componentRepositoryRoot) {
        assert componentConfiguration != null;
        assert componentRepositoryRoot != null;

        this.componentRepositoryRoot = componentRepositoryRoot;
        System.out.println(String.format("------------- Created component under test, repository root '%s'",
                this.componentRepositoryRoot.getAbsolutePath()));
        this.componentConfiguration = componentConfiguration;
        this.jbi = generateComponentJbi();
        copyOthersJbiResources();
        copyJbiResources(componentConfiguration.getResources(), getInstallRoot());
        this.component = createComponent(jbi.getComponent().getComponentClassName().getContent());
    }

    private final File getInstallRoot() {
        return new File(componentRepositoryRoot, "install");
    }

    private final File getWorkspaceRoot() {
        return new File(componentRepositoryRoot, "work");
    }

    /**
     * Generate the JBI descriptor from the current component configuration
     * 
     * @return The JBI descriptor
     */
    private Jbi generateComponentJbi() {

        final ComponentJbiGenerator componentJbiGenerator = new ComponentJbiGenerator(this.componentConfiguration,
                getInstallRoot());
        final File generatedJbiFile = componentJbiGenerator.generate();
        assert generatedJbiFile.exists();

        try {
            final FileInputStream fis = new FileInputStream(generatedJbiFile);
            try {
                return CDKJBIDescriptorBuilder.getInstance().buildJavaJBIDescriptor(fis);
            } catch (final JBIDescriptorException e) {
                throw new UncheckedException(e);
            } finally {
                IOHelper.close(fis);
            }
        } catch (final FileNotFoundException fnfe) {
            throw new UncheckedException(fnfe);
        }
    }

    public final String getComponentName() {
        return jbi.getComponent().getIdentification().getName();
    }

    private void copyOthersJbiResources() {

        final URL executionDirUrl = this.getClass().getClassLoader().getResource(JBI_DIR_NAME);
        assert executionDirUrl != null;

        try {
            if (executionDirUrl.getProtocol().equals("jar")) {

                final String executionDirStr = executionDirUrl.toString();
                final int executionJarEndIndex = executionDirStr.indexOf(".jar!") + ".jar".length();
                assert executionJarEndIndex != -1;

                final String executionJarStr = executionDirStr.substring("jar:file:".length(), executionJarEndIndex);
                final JarFile executionJar = new JarFile(executionJarStr);

                final File jbiDirExtractedFromComponentJar = new File(this.componentRepositoryRoot,
                        "jbiDirExtractedFromComponentJar");
                ZipHelper.extractEntriesFromArchive(executionJar, JBI_DIR_NAME + File.separator,
                        jbiDirExtractedFromComponentJar);
                this.copyAllJbiDirContentExceptJbiXmlFile(new File(jbiDirExtractedFromComponentJar, JBI_DIR_NAME));
            } else {
                final File executionDir = new File(executionDirUrl.toURI());
                this.copyAllJbiDirContentExceptJbiXmlFile(executionDir);
            }
        } catch (final Exception e) {
            throw new UncheckedException(e);
        }
    }

    private void copyAllJbiDirContentExceptJbiXmlFile(final File jbiDir) throws IOException {
        for (final File fileInJbiDir : jbiDir.listFiles()) {
            if ("jbi.xml".equals(fileInJbiDir.getName())) {
                // jbi.xml was generated just before, skip it.
                continue;
            }
            if (fileInJbiDir.isDirectory()) {
                FileUtils.copyDirectoryToDirectory(fileInJbiDir, this.getInstallRoot());
            } else {
                FileUtils.copyFileToDirectory(fileInJbiDir, this.getInstallRoot());
            }
        }
    }

    private void copyJbiResources(final URL[] resourceUrls, final File installationRoot) {
        for (final URL resourceUrl : resourceUrls) {
            try {
                if (resourceUrl.getProtocol().equals("jar")) {
                    // TODO : not supported, to be implemented... otherwise
                    // resource copying when resource in jar will not work
                    continue;
                }
                final File resourceFile = new File(resourceUrl.toURI());
                if (resourceFile.isDirectory()) {
                    FileUtils.copyDirectory(resourceFile, installationRoot);
                } else {
                    FileUtils.copyFileToDirectory(resourceFile, installationRoot);
                }
            } catch (final Exception e) {
                throw new UncheckedException(e);
            }
        }
    }

    private void copyWsdl(final URL serviceDescription, final File installationRoot) {
        try {
            final File resourceFile = new File(serviceDescription.toURI());
            FileUtils.copyFileToDirectory(resourceFile, installationRoot);
        } catch (final IOException | URISyntaxException e) {
            throw new UncheckedException(e);
        }
    }

    private AbstractComponent createComponent(final String componentClassName) {
        try {
            return (AbstractComponent) Class.forName(componentClassName).newInstance();
        } catch (final Exception e) {
            throw new UncheckedException(e);
        }
    }

    public void install() {
        this.mockComponentContext = new MockComponentContext(jbi.getComponent().getIdentification().getName(),
                this.component, this.getInstallRoot().getAbsolutePath(), this.getWorkspaceRoot().getAbsolutePath(),
                this.componentConfiguration.getLogger());
        final InstallationContext mockInstallationContext = new MockInstallationContext(this.getInstallRoot(),
                this.jbi, this.mockComponentContext, true);
        // jbi bootstrap
        try {
            @SuppressWarnings("unchecked")
            final Class<? extends JBIBootstrap> jbiBootstrapClass = (Class<? extends JBIBootstrap>) Class
                    .forName(this.jbi.getComponent().getBootstrapClassName());
            final JBIBootstrap jbiBootstrap = jbiBootstrapClass.newInstance();
            jbiBootstrap.init(mockInstallationContext);
            jbiBootstrap.onInstall();

        } catch (final Exception e) {
            throw new UncheckedException(e);
        }

        // jbi initialize
        try {
            this.component.init(mockInstallationContext.getContext());
            this.componentState = ComponentState.STOPPED;
        } catch (final JBIException jbie) {
            throw new UncheckedException(jbie);
        }
    }

    public boolean isInstalled() {
        return this.componentState == ComponentState.STOPPED || this.componentState == ComponentState.STARTED;
    }

    public void start() {
        try {
            this.component.start();
            this.componentState = ComponentState.STARTED;
        } catch (final JBIException e) {
            throw new UncheckedException("Impossible to start the wrapped component", e);
        }
    }

    public boolean isStarted() {
        return this.componentState == ComponentState.STARTED;
    }

    public void stop() {
        try {
            this.component.stop();
            this.componentState = ComponentState.STOPPED;
        } catch (final JBIException e) {
            throw new UncheckedException("Impossible to stop the component under test", e);
        }
    }

    public void shutdown() {
        try {
            if (this.isStarted()) {
                this.stop();
            }

            // To shutdown the component, all SUs must be undeployed
            final AbstractServiceUnitManager serviceUnitManager = this.component.getServiceUnitManager();
            for (final ServiceUnitDataHandler suDH : serviceUnitManager.getServiceUnitDataHandlers()) {
                final String serviceUnitName = suDH.getName();
                serviceUnitManager.stop(serviceUnitName);
                serviceUnitManager.undeploy(serviceUnitName, suDH.getInstallRoot());
            }
            this.component.shutDown();
            this.componentState = ComponentState.NOT_INSTALLED;
        } catch (final JBIException e) {
            throw new UncheckedException("Impossible to stop the component under test", e);
        }
    }

    public void deployService(final String suName, final ServiceConfiguration serviceConfiguration)
            throws DeploymentException {

        final AbstractServiceUnitManager serviceUnitManager = this.component.getServiceUnitManager();
        final File serviceUnitDir = this.getServiceConfigurationInstallDir(suName);

        if (serviceConfiguration.getServiceDescription() != null) {
            this.copyWsdl(serviceConfiguration.getServiceDescription(), serviceUnitDir);
        }
        this.copyJbiResources(serviceConfiguration.getResources(), serviceUnitDir);

        final ServiceUnitJbiGenerator serviceUnitJbiGenerator = new ServiceUnitJbiGenerator(serviceConfiguration,
                serviceUnitDir);
        boolean multiservice = false;
        serviceUnitJbiGenerator.generate(multiservice);

        final String result = serviceUnitManager.deploy(suName, serviceUnitDir.getAbsolutePath());
        if (result.contains("<task-result>SUCCESS</task-result>")) {
            serviceUnitManager.init(suName, serviceUnitDir.getAbsolutePath());
            serviceUnitManager.start(suName);
        } else {
            throw new DeploymentException("Impossible to install the specified service: " + result);
        }
        final Collection<ServiceEndpoint> endpoints = serviceUnitManager.getEndpointsForServiceUnit(suName);
        if (serviceConfiguration.getServiceType() == ServiceType.PROVIDE) {
            // we deployed only one provides!
            Assert.assertTrue(endpoints.size() == 1);
        } else {
            // we deployed no provides
            Assert.assertTrue(endpoints.size() == 0);
        }
        // if it was autogenerated, let's get the generated name
        if (serviceConfiguration.getEndpointName().equals("autogenerate")) {
            serviceConfiguration.setEndpointName(endpoints.iterator().next().getEndpointName());
        }
    }

    public File getServiceConfigurationInstallDir(final String suName) {
        final File serviceUnitsInstallRoot = new File(this.componentRepositoryRoot, "service-units");
        return new File(serviceUnitsInstallRoot, suName);
    }

    public void undeployService(final String suName) {
		try {
            final AbstractServiceUnitManager serviceUnitManager = (AbstractServiceUnitManager) this.component
					.getServiceUnitManager();

            final ServiceUnitDataHandler serviceUnitDataHandler = serviceUnitManager.getSUDataHandler(suName);
			serviceUnitManager.stop(suName);
			serviceUnitManager.shutDown(suName);
			serviceUnitManager.undeploy(suName,	serviceUnitDataHandler.getInstallRoot());
        } catch (final DeploymentException e) {
			throw new UncheckedException(
					"Impossible to uninstall the specified service", e);
		}
    }

	public void undeployAllServices() {
		try {
            final AbstractServiceUnitManager serviceUnitManager = this.component.getServiceUnitManager();
	        
            for (ServiceUnitDataHandler serviceUnitDataHandler : serviceUnitManager.getServiceUnitDataHandlers()) {
                final String suName = serviceUnitDataHandler.getName();
				serviceUnitManager.stop(suName);
				serviceUnitManager.shutDown(suName);
				serviceUnitManager.undeploy(suName,	serviceUnitDataHandler.getInstallRoot());
	        }
        } catch (final DeploymentException e) {
			throw new UncheckedException(
					"Impossible to uninstall all the services", e);
		}
	}
	
    public boolean isServiceDeployed(final String suName) {
        return component.getServiceUnitManager().getSUDataHandler(suName) != null;
    }

    public void pushSynchronouslyRequestToProvider(final RequestMessage message) {
        assert message instanceof AbstractWrappedToComponentMessage;

        try {
            final PetalsMessageExchange msgExchange = (PetalsMessageExchange) message.getMessageExchange();

            final Semaphore sem = new Semaphore(0);
            final JBIProcessorManager processorManager = this.component.getProcessorManager();
            final JBIProcessorManagerObserver jbiProccesorManagerObserver = new JBIProcessorManagerObserver() {
                @Override
                public void beforeProcessingMessage(final Exchange exchange) {
                    // do nothing
                }

                @Override
                public void afterProcessingMessage(final Exchange exchange) {
                    if (exchange.getExchangeId().equals(msgExchange.getExchangeId())) {
                        // this will unblock the execution of pushSync until
                        // the message has been processed by the provider
                        sem.release();
                    }
                }
            };

            processorManager.addObserver(jbiProccesorManagerObserver);

            this.mockComponentContext.getDeliveryChannel().enqueueIncomingExchange(msgExchange);

            sem.acquire();

            processorManager.deleteObserver(jbiProccesorManagerObserver);
        } catch (final InterruptedException e) {
            throw new UncheckedException("Impossible to send a message from kernel", e);
        }
    }

    private final void pushMessageInDeliveryChannel(final Message message) {
        final PetalsMessageExchange msgExchange = (PetalsMessageExchange) message.getMessageExchange();

        this.mockComponentContext.getDeliveryChannel().enqueueIncomingExchange(msgExchange);
    }
    
    public void pushRequestToProvider(final RequestMessage message) {

        assert message instanceof AbstractWrappedToComponentMessage;

        this.pushMessageInDeliveryChannel(message);
    }
    
    public void pushStatusToProvider(final ResponseMessage message) {

        assert message instanceof AbstractWrappedToComponentMessage;

        this.pushMessageInDeliveryChannel(message);
    }
    
    public void pushResponseToConsumer(final ResponseMessage message) {

        assert message instanceof AbstractWrappedToComponentMessage;

        final FlowAttributes flowAttributes = PetalsExecutionContext.getFlowAttributes();
        String flowInstanceId = null;
        String flowStepId = null;
        if (flowAttributes != null) {
            flowInstanceId = flowAttributes.getFlowInstanceId();
            flowStepId = flowAttributes.getFlowStepId();
        }
        final AbstractFlowLogData flowLogData;
        final Exchange exchange = new ExchangeImpl(message.getMessageExchange());
        if (!exchange.isErrorStatus() && exchange.getFault() == null) {
            flowLogData = new ProvideFlowStepEndLogData(flowInstanceId, flowStepId, message.getMessageExchange());
        } else {
            final String failureMessage;
            if (exchange.isErrorStatus()) {
                Exception error = exchange.getError();
                assert error != null;
                if (error.getMessage() != null) {
                    failureMessage = error.getMessage();
                } else {
                    failureMessage = error.getStackTrace()[0].toString();
                }
            } else {
                failureMessage = "BUSINESS ERROR MESSAGE";
            }

            flowLogData = new ProvideFlowStepFailureLogData(flowInstanceId, flowStepId, failureMessage,
                    message.getMessageExchange());
        }

        this.component.getLogger().log(Level.MONIT, "", flowLogData);

        this.pushMessageInDeliveryChannel(message);
    }

    public ResponseMessage pollResponseFromProvider(final long timeout) {
        final PetalsMessageExchange msgExchange = this.mockComponentContext.getDeliveryChannel()
                .dequeueExchangeToClient(timeout);

        // if timeout
        if (msgExchange == null) {
            return null;
        }

        assert!MEPConstants.IN_ONLY_PATTERN.equals(msgExchange
                .getPattern()) : "Polling response from provider while processing an IN-ONLY exchange is meaningless.";

        return new WrappedResponseFromProviderMessage(msgExchange);
    }

    public ResponseMessage pollStatusFromProvider(final long timeout) {
        final PetalsMessageExchange msgExchange = this.mockComponentContext.getDeliveryChannel()
                .dequeueExchangeToClient(timeout);

        // if timeout
        if (msgExchange == null) {
            return null;
        }

        assert!MEPConstants.IN_OUT_PATTERN.equals(msgExchange
                .getPattern()) : "Polling status from provider while processing an in IN-OUT exchange is meaningless.";

        return new WrappedResponseFromProviderMessage(msgExchange);
    }
    
    public ResponseMessage pollStatusFromConsumer(final long timeout) {
        final PetalsMessageExchange msgExchange = this.mockComponentContext.getDeliveryChannel()
                .dequeueExchangeToClient(timeout);

        // if timeout
        if (msgExchange == null) {
            return null;
        }

        assert!MEPConstants.IN_ONLY_PATTERN.equals(msgExchange
                .getPattern()) : "Polling status from consumer while processing an in IN-ONLY exchange is meaningless.";

        return new WrappedResponseFromProviderMessage(msgExchange);
    }

    public void clearResponsesFromProvider() {
        this.mockComponentContext.getDeliveryChannel().clearExchangesToClient();
    }

    public int getResponsesFromProviderCount() {
        return this.mockComponentContext.getDeliveryChannel().getExchangesToClientCount();
    }

    public RequestMessage pollRequestFromConsumer(final long timeout) {
        final PetalsMessageExchange msgExchange = this.mockComponentContext.getDeliveryChannel()
                .dequeueExchangeToServer(timeout);

        // if timeout
        if (msgExchange == null) {
            return null;
        }

        if (msgExchange.getStatus().equals(ExchangeStatus.ACTIVE)) {
            final FlowAttributes flowAttributes = FlowAttributesExchangeHelper.getFlowAttributes(msgExchange);
            if (flowAttributes != null) {
                PetalsExecutionContext.putFlowAttributes(flowAttributes);
                PetalsExecutionContext.nextFlowStepId();
            }

            String edptName = "";
            QName serviceName = null;
            if (msgExchange.getEndpoint() != null) {
                edptName = msgExchange.getEndpoint().getEndpointName();
                serviceName = msgExchange.getEndpoint().getServiceName();
            }
            final Properties mdcProperties = ExecutionContext.getProperties();

            this.component.getLogger().log(Level.MONIT, "",
                    new ProvideFlowStepBeginLogData(mdcProperties.getProperty(FLOW_INSTANCE_ID_PROPERTY_NAME),
                            mdcProperties.getProperty(FLOW_STEP_ID_PROPERTY_NAME),
                            StringHelper.nonNullValue(msgExchange.getInterfaceName()),
                            StringHelper.nonNullValue(serviceName),
                            StringHelper.nonNullValue(msgExchange.getOperation()), edptName,
                            mdcProperties.getProperty(FLOW_PREVIOUS_STEP_ID_PROPERTY_NAME), msgExchange));
        }

        return new WrappedRequestFromConsumerMessage(msgExchange);
    }

    public void clearRequestsFromConsumer() {
        this.mockComponentContext.getDeliveryChannel().clearExchangesToServer();
    }

    public int getRequestsFromConsumerCount() {
        return this.mockComponentContext.getDeliveryChannel().getExchangesToServerCount();
    }

    public void clearIncomingMessages() {
        this.mockComponentContext.getDeliveryChannel().clearIncomingExchanges();
    }

    public String getNativeEndpointName(final QName service) {
        final Description nativeDescription = this.component.getNativeWsdl().getDescription();
        return nativeDescription.getService(service).getEndpoints().get(0).getName();
    }
}
