/**
 * Copyright (c) 2011-2012 EBM WebSourcing, 2012-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.junit.impl;

import static org.ow2.petals.component.framework.junit.JbiConstants.JBI_NAMESPACE_URI;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.URL;

import javax.xml.namespace.QName;
import javax.xml.transform.Result;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import com.ebmwebsourcing.easycommons.lang.UncheckedException;
import com.ebmwebsourcing.easycommons.xml.DOMHelper;
import com.ebmwebsourcing.easycommons.xml.Transformers;

/**
 * @author Marc Jambert - EBM WebSourcing
 */
public final class ComponentJbiGenerator {

	private final DOMSource jbiSource;

	private final ComponentConfiguration componentConfiguration;

	private final File installationRoot;

    public ComponentJbiGenerator(final ComponentConfiguration componentConfiguration, final File installationRoot) {
		assert componentConfiguration != null;
		assert installationRoot != null;
        this.jbiSource = this.readDefaultJbiDescriptor();
		this.installationRoot = installationRoot;
		this.componentConfiguration = componentConfiguration;
	}

    /**
     * Read the default JBI descriptor available on the classloader of the current thread, and identified by the
     * resource 'jbi/jbi.xml'
     * 
     * @return The default JBI descriptor
     */
    private DOMSource readDefaultJbiDescriptor() {

        final URL defaultComponentJbiUrl = ComponentJbiGenerator.class.getClassLoader().getResource("jbi/jbi.xml");
        return DOMHelper.parseAsDOMSource(defaultComponentJbiUrl);
	}

    /**
     * Generate the component JBI descriptor with the component configuration previously provided.
     */
	public File generate() {

        this.replaceJbiParametersWithComponentConfigurationParameters();

        // dump JBI descriptor back to installationRoot.
        final File installationRootMetaInf = new File(this.installationRoot, "META-INF");
        installationRootMetaInf.mkdirs();
        final File jbiDumpTarget = new File(installationRootMetaInf, "jbi.xml");
        try {
            this.dumpModifiedJbi(jbiSource, jbiDumpTarget);
            return jbiDumpTarget;
        } catch (final IOException e) {
            throw new UncheckedException(e);
        }

	}

    /**
     * Write a JBI descriptor given as {@link DOMSource} into the given file.
     * 
     * @param jbiSource
     *            A JBI descriptor
     * @param jbiDumpTarget
     *            The file in which the JBI descriptor will be dumped
     * @throws IOException
     */
    private void dumpModifiedJbi(final DOMSource jbiSource, final File jbiDumpTarget) throws IOException {
        final FileOutputStream fos = new FileOutputStream(jbiDumpTarget);
        try {
            final Result result = new StreamResult(fos);
            final Transformer transformer = Transformers.takeTransformer();
            try {
                transformer.transform(jbiSource, result);
            } catch (final TransformerException te) {
                throw new UncheckedException(te);
            } finally {
                Transformers.releaseTransformer(transformer);
            }
        } finally {
            fos.close();
        }
	}

    private void replaceJbiParameterWithComponentConfigurationParameter(final QName parameterQName,
            final String componentConfigurationParameterValue) {
        assert parameterQName != null;

        final Document jbiDocument = (Document) this.jbiSource.getNode();
        assert jbiDocument != null;

        final NodeList parameterNodeList = jbiDocument.getElementsByTagNameNS(parameterQName.getNamespaceURI(),
                parameterQName.getLocalPart());
        final int parameterNodeListLength = parameterNodeList.getLength();
        assert parameterNodeListLength <= 1;

        final boolean isTagFound = parameterNodeListLength == 1;
        if (isTagFound) {
            final Node parameterNode = parameterNodeList.item(0);
            parameterNode.setTextContent(componentConfigurationParameterValue);
        } else {
            final Element parameterNode = jbiDocument.createElementNS(parameterQName.getNamespaceURI(),
                    parameterQName.getLocalPart());
            parameterNode.setTextContent(componentConfigurationParameterValue);
            final NodeList componentNodeList = jbiDocument.getElementsByTagNameNS(JBI_NAMESPACE_URI, "");
            final int componentNodeListLength = componentNodeList.getLength();
            assert componentNodeListLength == 1;

            final Node componentNode = componentNodeList.item(0);
            componentNode.appendChild(parameterNode);
        }

    }

    private void replaceJbiParametersWithComponentConfigurationParameters() {
        this.replaceJbiParameterWithComponentConfigurationParameter(new QName(JBI_NAMESPACE_URI, "name"),
                this.componentConfiguration.getConfigurationName());
        for (final QName parameterName : this.componentConfiguration.getParameterNames()) {
            replaceJbiParameterWithComponentConfigurationParameter(parameterName,
                    this.componentConfiguration.getParameter(parameterName));
        }
    }
}
