/**
 * Copyright (c) 2014-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.component.framework.junit.rule;

import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.logging.Handler;
import java.util.logging.LogManager;

import javax.jbi.management.DeploymentException;
import javax.xml.namespace.QName;

import org.apache.commons.io.FileUtils;
import org.junit.Rule;
import org.junit.rules.ExternalResource;
import org.junit.rules.TemporaryFolder;
import org.ow2.petals.commons.log.Level;
import org.ow2.petals.component.framework.junit.Component;
import org.ow2.petals.component.framework.junit.RequestMessage;
import org.ow2.petals.component.framework.junit.ResponseMessage;
import org.ow2.petals.component.framework.junit.impl.ComponentConfiguration;
import org.ow2.petals.component.framework.junit.impl.ComponentInstanceImpl;
import org.ow2.petals.component.framework.junit.impl.ServiceConfiguration;
import org.ow2.petals.component.framework.junit.impl.mock.MockComponentContext;
import org.ow2.petals.component.framework.junit.impl.mock.MockServiceEndpoint;
import org.ow2.petals.log.handler.PetalsFileHandler;

import com.ebmwebsourcing.easycommons.lang.UncheckedException;

/**
 * The {@link ComponentUnderTest} {@link Rule} allows creation of component instance of the current component project.
 * It is guaranteed to be deleted when the test method finishes (whether it passes or fails):
 * 
 * <pre>
 * public class MyTestSuite {
 *  &#064;Rule
 *  public ComponentUnderTest componentUnderTest = new ComponentUnderTest(...);
 * 
 *  &#064;Test
 *  public void test...() throws IOException {
 *      componentUnderTest.start();
 *      componentUnderTest.installService(...);
 *      ...
 *  }
 * }
 * </pre>
 * 
 * If a file logging.properties is present in the classpath, it is used for the logs.
 */
public class ComponentUnderTest extends ExternalResource implements Component {

    static {

        // let's initialize logging stuffs, it has to be done here to happen before the rest.
        Level.initialize();

        final InputStream logConfig = ComponentUnderTest.class.getResourceAsStream("/logging.properties");
        if (logConfig != null) {
            try {
                LogManager.getLogManager().readConfiguration(logConfig);
            } catch (final SecurityException | IOException e) {
                fail(e.getMessage());
            }
        }
    }

    /**
     * A temporary folder that will be used to define the root of the file-system used by the component instance.
     */
    private final TemporaryFolder temporaryFolder = new TemporaryFolder();

    /**
     * The component instance configuration
     */
    private final ComponentConfiguration componentConfiguration;

    /**
     * The component instance
     */
    private ComponentInstanceImpl componentInst = null;

    /**
     * Flag controlling the component must be installed on setup
     */
    private final boolean componentMustBeInstalled;

    /**
     * Flag controlling if the component must be started on setup
     */
    private final boolean componentMustBeStarted;

    /**
     * The directory where the component is installed
     */
    private File componentInstallationDir = null;

    /**
     * Service configurations factory creating the service configuration to deploy when the component will be
     * installed/started
     */
    private final Map<String, ServiceConfigurationFactory> serviceCfgFactoryRegisteredForDeploiement = new HashMap<String, ServiceConfigurationFactory>();

    /**
     * Native service configurations factory creating the service configuration to deploy when the component will be
     * installed/started
     */
    private final Map<String, NativeServiceConfigurationFactory> nativeServiceCfgFactoryRegisteredForDeploiement = new HashMap<String, NativeServiceConfigurationFactory>();

    /**
     * Service configurations deployed
     */
    private final Map<String, ServiceConfiguration> serviceConfigurations = new HashMap<String, ServiceConfiguration>();

    /**
     * The files to add in the component file hierarchy (hack needed by Petals BC SOAP to load the addressing module)
     */
    private final Map<String, String> filesToAdd = new HashMap<String, String>();

    /**
     * 
     * Create a component under test. It is automatically installed and started
     * 
     * @param componentConfiguration
     *            The component configuration
     */

    public ComponentUnderTest() {
        this(new ComponentConfiguration("UnderTestWithDefaultConf"), true, true);
    }

    /**
     * 
     * @param componentConfiguration
     *            The component configuration
     * @param installesComponent
     *            Flag controlling the component must be installed on setup
     * @param startsComponent
     *            Flag controlling if the component must be started on setup
     */
    public ComponentUnderTest(final ComponentConfiguration componentConfiguration, final boolean installesComponent,
            final boolean startsComponent) {
        this.componentConfiguration = componentConfiguration;
        assertTrue("Unable to start a component not initialized", (startsComponent && installesComponent)
                || !startsComponent);
        this.componentMustBeInstalled = installesComponent;
        this.componentMustBeStarted = startsComponent;
    }

    @Override
    protected void before() throws Throwable {
        this.temporaryFolder.create();
        
        this.componentInstallationDir = this.temporaryFolder.newFolder();

        System.setProperty(PetalsFileHandler.PETALS_LOG_DIR_PROPERTY_NAME, new File(this.componentInstallationDir,
                "log").getAbsolutePath());

        for (final Entry<String, String> fileToAddEntry : this.filesToAdd.entrySet()) {
            final File srcFile = new File(fileToAddEntry.getKey());
            final File dstFile = new File(new File(this.componentInstallationDir, fileToAddEntry.getValue()),
                    srcFile.getName());
            FileUtils.copyFile(srcFile, dstFile);
        }

        this.componentInst = new ComponentInstanceImpl(this.componentConfiguration, this.componentInstallationDir);
        if (this.componentMustBeInstalled) {
            this.componentInst.install();

            for (final Entry<String, ServiceConfigurationFactory> serviceToDeploy : this.serviceCfgFactoryRegisteredForDeploiement
                    .entrySet()) {
                this.deployService(serviceToDeploy.getKey(), serviceToDeploy.getValue());
            }

            for (final Entry<String, NativeServiceConfigurationFactory> serviceToDeploy : this.nativeServiceCfgFactoryRegisteredForDeploiement
                    .entrySet()) {
                this.registerNativeService(serviceToDeploy.getKey(), serviceToDeploy.getValue());
            }

            if (this.componentMustBeStarted) {
                this.start();
            }
        }
    }

    @Override
    protected void after() {

        if (this.isStarted()) {
            this.stop();
        }

        if (this.isInstalled()) {
            this.undeployAllServices();
            this.shutdown();
        }

        // clear the queues
        componentInst.clearRequestsFromConsumer();
        componentInst.clearResponsesFromProvider();
        componentInst.clearIncomingMessages();

        System.clearProperty(PetalsFileHandler.PETALS_LOG_DIR_PROPERTY_NAME);

        this.temporaryFolder.delete();
    }

    @Override
    public ComponentUnderTest addLogHandler(final Handler logHandler) {
        this.componentConfiguration.getLogger().addHandler(logHandler);
        return this;
    }

    @Override
    public String getComponentName() {
        return this.componentInst.getComponentName();
    }

    @Override
    public File getBaseDirectory() {
        return this.componentInstallationDir;
    }

    @Override
    public boolean isInstalled() {
        return this.componentInst != null && this.componentInst.isInstalled();
    }

    @Override
    public void start() {
        this.componentInst.start();
    }

    @Override
    public boolean isStarted() {
        return this.componentInst != null && this.componentInst.isStarted();
    }

    @Override
    public void stop() {
        this.componentInst.stop();
    }

    @Override
    public void shutdown() {
        this.componentInst.shutdown();
    }

    @Override
    public ComponentUnderTest registerServiceToDeploy(final String serviceCfgId,
            final ServiceConfigurationFactory serviceConfigurationFactory) {
        assertNull("Service configuration already registered",
                this.serviceCfgFactoryRegisteredForDeploiement.get(serviceCfgId));
        this.serviceCfgFactoryRegisteredForDeploiement.put(serviceCfgId, serviceConfigurationFactory);
        return this;
    }

    @Override
    public void deployService(final String serviceCfgId, final ServiceConfigurationFactory serviceConfigurationFactory)
            throws DeploymentException {
        ServiceConfiguration serviceConfiguration;
        try {
            serviceConfiguration = serviceConfigurationFactory.create();
        } catch (final Exception e) {
            throw new DeploymentException(e);
        }
        this.deployService(serviceCfgId, serviceConfiguration);
    }

    @Override
    public void deployService(String suName, ServiceConfiguration serviceConfiguration) throws DeploymentException {
        this.serviceConfigurations.put(suName, serviceConfiguration);
        this.componentInst.deployService(suName, serviceConfiguration);
    }

    @Override
    public boolean isServiceDeployed(final String suName) {
        return this.componentInst.isServiceDeployed(suName);
    }

    @Override
    public void undeployService(final String serviceConfigurationId) {
        final ServiceConfiguration serviceConfiguration = this.serviceConfigurations.remove(serviceConfigurationId);
        if (serviceConfiguration != null) {
            this.componentInst.undeployService(serviceConfigurationId);
        }
    }

    @Override
    public void undeployAllServices() {
        this.componentInst.undeployAllServices();
        this.serviceConfigurations.clear();
    }

    @Override
    public ComponentUnderTest registerNativeServiceToDeploy(final String serviceCfgId,
            final NativeServiceConfigurationFactory serviceConfigurationFactory) {
        assertNull("Service configuration already registered",
                this.nativeServiceCfgFactoryRegisteredForDeploiement.get(serviceCfgId));
        this.nativeServiceCfgFactoryRegisteredForDeploiement.put(serviceCfgId, serviceConfigurationFactory);
        return this;
    }

    @Override
    public void registerNativeService(final String serviceCfgId,
            final NativeServiceConfigurationFactory serviceConfigurationFactory) {
        assertNull("Service configuration already registered", this.serviceConfigurations.get(serviceCfgId));
        final ServiceConfiguration serviceConfiguration = serviceConfigurationFactory.create(this.componentInst
                .getNativeEndpointName(serviceConfigurationFactory.getNativeService()));
        this.serviceConfigurations.put(serviceCfgId, serviceConfiguration);
    }

    @Override
    public ComponentUnderTest registerExternalServiceProvider(final QName serviceName, final String endpointName) {
        MockComponentContext.activateEndpoint(new MockServiceEndpoint(endpointName, serviceName));
        return this;
    }

    @Override
    public void pushRequestToProvider(final RequestMessage message) {
        this.componentInst.pushRequestToProvider(message);
    }

    @Override
    public void pushStatusToProvider(final ResponseMessage message) {
        this.componentInst.pushStatusToProvider(message);
    }

    @Override
    public ResponseMessage pollResponseFromProvider() {
        return this.componentInst.pollResponseFromProvider(DEFAULT_TIMEOUT);
    }

    @Override
    public ResponseMessage pollResponseFromProvider(long timeout) {
        return this.componentInst.pollResponseFromProvider(timeout);
    }

    @Override
    public ResponseMessage pollStatusFromProvider() {
        return this.componentInst.pollStatusFromProvider(DEFAULT_TIMEOUT);
    }

    @Override
    public ResponseMessage pollStatusFromProvider(long timeout) {
        return this.componentInst.pollStatusFromProvider(timeout);
    }
    
    @Override
    public ResponseMessage pollStatusFromConsumer() {
        return this.componentInst.pollStatusFromConsumer(DEFAULT_TIMEOUT);
    }

    @Override
    public ResponseMessage pollStatusFromConsumer(long timeout) {
        return this.componentInst.pollStatusFromConsumer(timeout);
    }

    @Override
    public void clearResponsesFromProvider() {
        this.componentInst.clearResponsesFromProvider();
    }

    @Override
    public int getResponsesFromProviderCount() {
        return this.componentInst.getResponsesFromProviderCount();
    }

    @Override
    public RequestMessage pollRequestFromConsumer() {
        return this.componentInst.pollRequestFromConsumer(DEFAULT_TIMEOUT);
    }

    @Override
    public RequestMessage pollRequestFromConsumer(long timeout) {
        return this.componentInst.pollRequestFromConsumer(timeout);
    }

    @Override
    public void clearRequestsFromConsumer() {
        this.componentInst.clearRequestsFromConsumer();
    }

    @Override
    public int getRequestsFromConsumerCount() {
        return this.componentInst.getRequestsFromConsumerCount();
    }

    @Override
    public void pushResponseToConsumer(final ResponseMessage message) {
        this.componentInst.pushResponseToConsumer(message);
    }

    @Override
    public void pushSynchronouslyRequestToProvider(final RequestMessage message) {
        this.componentInst.pushSynchronouslyRequestToProvider(message);
    }

    @Override
    public ServiceConfiguration getServiceConfiguration(final String suName) {
        return this.serviceConfigurations.get(suName);
    }

    @Override
    public String getNativeEndpointName(final QName service) {
        return this.componentInst.getNativeEndpointName(service);
    }

    @Override
    public File getServiceConfigurationInstallDir(final String suName) {
        return this.componentInst.getServiceConfigurationInstallDir(suName);
    }

    /**
     * Set a parameter of the JBI descriptor of component under test. If an error occurs, an {@link UncheckedException}
     * is thrown.
     * 
     * @param paramName
     *            The parameter name
     * @param value
     *            The parameter value
     */
    public ComponentUnderTest setParameter(final QName paramName, final String value) {
        this.componentConfiguration.setParameter(paramName, value);
        return this;
    }

    /**
     * Set a parameter of the JBI descriptor of component under test. If an error occurs, an {@link UncheckedException}
     * is thrown.
     * 
     * @param paramName
     *            The parameter name
     * @param paramGenerator
     *            The parameter generator used to generate the content of the value
     */
    public ComponentUnderTest setParameter(final QName paramName, final ParameterGenerator paramGenerator) {
        try {
            return this.setParameter(paramName, paramGenerator.generate());
        } catch (final Exception e) {
            throw new UncheckedException(e);
        }
    }

    /**
     * Add a file in the component under test file hierachy (Hack required by Petals BC SOAP to engage its module
     * 'addressing'
     * 
     * @param fileToAdd
     *            The file to add
     * @param destInComponentDir
     *            The destination path relative to the root directory of the component under test
     */
    public ComponentUnderTest addFile(final String fileToAdd, final String destInComponentDir) {
        this.filesToAdd.put(fileToAdd, destInComponentDir);
        return this;
    }
}
